import contextlib
import datetime
import functools
import sys
import time

from fattoush import world
from selenium.common.exceptions import WebDriverException


@contextlib.contextmanager
def _screenshot_after(step):
    """
    Ensure that a screenshot is taken after the decorated step definition
    is run.
    """
    parent = step.parent
    parent_name = getattr(parent, 'name', None)
    if parent_name is None:  # Must be a background
        parent_name = parent.feature.name

    filename = (
        "logs/{datetime:{dfmt}_{tfmt}}{parent_name}.{step.sentence}.png"
        .format(
            step=step,
            parent_name=parent_name,
            datetime=datetime.datetime.now(),
            dfmt='%Y%m%d',
            tfmt='%H%M%S.%f',
        )
    )

    try:
        yield
    except:
        exc_type, exc_value, exc_tb = sys.exc_info()

        time.sleep(1)

        browser = world.per_scenario.get('browser')

        if browser is not None:
            try:
                browser.get_screenshot_as_file(filename)
            except Exception as ex:
                print(
                    "could not capture screen shot to {}:\n{}"
                    .format(filename, ex)
                )
            else:
                print("captured screen shot to {}".format(filename))

        raise exc_type, exc_value, exc_tb
    else:
        browser = world.per_scenario.get('browser')
        if browser is None:
            return

        try:
            if browser.is_sauce:
                browser.get_screenshot_as_png()
            else:
                browser.get_screenshot_as_file(filename)
        except WebDriverException:
            pass


def screenshot(fn):
    @functools.wraps(fn)
    def _inner(step, *args, **kwargs):
        with _screenshot_after(step):
            return fn(step, *args, **kwargs)
    return _inner



# -*- coding: utf-8 -*-
# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

from fattoush import step, Driver
from nose import tools
from time import sleep

@step(u'I open "([^"]*)" in my web browser')
def i_open_url_in_my_web_browser(step, url):
    wd = Driver.instance(step)
    wd.get(url)

@step(u'I expect the url to start with "([^"]*)"')
def i_expect_the_url_to_become_url(step, url):
    wd = Driver.instance(step)
    tools.assert_equal(wd.current_url[:len(url)], url)


@step(u'I type "([^"]*)" into the search box')
def i_type_search_term_into_the_search_box(step, search_term):
    wd = Driver.instance(step)
    search_box = wd.find_element_by_id('gbqfq')
    search_box.send_keys(search_term)

@step(u'I submit the search')
def i_submit_the_search(step):
    wd = Driver.instance(step)
    search_box = wd.find_element_by_id('gbqfq')
    search_box.send_keys('\n')

@step(u'I expect the top result to contain the string "([^"]*)"')
def i_expect_the_top_result_to_contain_the_string_search_term(step,
                                                              search_term):
    wd = Driver.instance(step)
    assert isinstance(wd, Driver)
    for i in range(5):
        try:
            results = wd.find_elements_by_class_name('rc')
            top_result = results[0]
            break
        except IndexError:
            sleep(1)
            continue
    else:
        raise
    tools.assert_true(search_term in top_result.text,
                      u'could not find {0} in {1}'.format(
                          search_term, top_result.text))



"""
Fattoush
========

Fattoush is a package that combines lettuce, webdriver and sauce to make
a tasty UI testing salad.


Running
-------

Fattoush provides its own test runner which invokes lettuce such that it
runs in saucelabs, with support of parallel test runs.

    $ fattoush --parallel=webdriver


Usage
-----

Just write your lettuce tests as normal,
but in your terrain file add

    from fattoush import hooks

You will now find that in each step you are able to add

    from fattoush import Driver

This returns a subclass of WebDriver with a class method `instance()`
which takes the current step or feature and returns the active WebDriver
instance.

License
-------
(c) 2014 Mind Candy Ltd. All Rights Reserved.
Licensed under the MIT License; you may not use this file except in compliance with the License.
You may obtain a copy of the License at http://opensource.org/licenses/MIT.
"""

# Bringing the step and world names into this namespace makes it
# easier to replace them at some future point.
#noinspection PyUnresolvedReferences
from lettuce import step as step_, world

from .driver import Driver
from .config import VERSION
from . import decorators


def step(regex):
    define = step_(regex)

    def _inner(fn):
        decorated = decorators.step.screenshot(fn)
        return define(decorated)

    return _inner



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

"""
The driver class shall be a subclass of WebDriver with a classmethod
`instance()` which takes the current step or feature and returns
the active WebDriver instance.
"""

from selenium.webdriver import Remote
from lettuce.core import Step, Scenario
from lettuce import world
from .sauce import Sauce, Local


class Driver(Remote):

    @classmethod
    def _scenario(cls, step_or_scenario):
        if isinstance(step_or_scenario, Scenario):
            return step_or_scenario.name

        if isinstance(step_or_scenario, Step):
            step = step_or_scenario

            if getattr(step, 'scenario', None) is not None:
                return step.scenario.name

            if getattr(step, 'background', None) is not None:
                return repr(step)

        raise TypeError("{0} is not an instance of {1} or {2}"
                        .format(step_or_scenario, Step, Scenario))

    @classmethod
    def instance(cls, step_or_scenario):
        """
        :rtype : Driver
        """

        if 'browser' not in world.per_scenario:
            world.per_scenario['browser'] = cls(
                world.fattoush,
                cls._scenario(step_or_scenario),
            )

        return world.per_scenario['browser']

    @classmethod
    def got_instance(cls):
        return 'browser' in world.per_scenario

    @classmethod
    def kill_instance(cls):
        browser = world.per_scenario.pop('browser', None)

        if browser is not None:
            browser.quit()

    def __init__(self, config, name):
        """
        :param config: fattoush.config.FattoushConfig
        :param scenario: Scenario
        """
        self.fattoush_config = config

        super(Driver, self).__init__(
            config.command_executor,
            config.desired_capabilities(name),
        )

        # For saucelabs you need a user name and an auth key
        self.is_sauce = "user" in config.server

        cls = Sauce if self.is_sauce else Local
        self.sauce = cls(config=config, browser=self)



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

"""
Functions and classes for taking commandline arguments and forming an
object which will give you a set of FattoushConfig objects, each of
which contains all the information necessary to run lettuce against a
specified webdriver configuration, whither in saucelabs or local.
"""
import json
from multiprocessing import Process
from os import environ, path
from lettuce import world
from ..runner.parsing import parse_args
import urlparse
from jsonschema import validate
from .config import FattoushConfig


class FattoushConfigGroup(object):

    @staticmethod
    def config_from_env():
        """
        Joyfully SauceConnect presents completely different
        environmental variables based on whether you are
        running against one saucelabs session or several.

        This function will return a list of different session
        configurations whatever the case - empty if there are
        none, a list with only one item if there is only one,
        or a list of multiple items if there are many. It's
        almost as if this would have been the sensible way
        for them to do it too.

        The only part which is different in the singular case is
        the missing 'os' key.

        in the multiple case the json string is documented to be
        of the format as follows:

        [
            {
                "platform":"LINUX",
                "os":"Linux",
                "browser":"firefox",
                "url":"sauce-ondemand:?os=Linux&
                                       browser=firefox&
                                       browser-version=16",
                "browser-version":"16"
            },
            {
                "platform":"VISTA",
                "os":"Windows 2008",
                "browser":"iexploreproxy",
                "url":"sauce-ondemand:?os=Windows 2008&
                                       browser=iexploreproxy&
                                       browser-version=9",
                "browser-version":"9"
            }
        ]
        """

        try:
            json_data = environ.get("SAUCE_ONDEMAND_BROWSERS")
            browsers = json.loads(json_data)
        except (ValueError, TypeError):
            url = environ.get("SELENIUM_DRIVER")
            try:
                query = urlparse.urlparse(url).query
                parsed = urlparse.parse_qs(query)
            except AttributeError:
                parsed = {}
            browsers = [
                {
                    "platform": environ.get("SELENIUM_PLATFORM"),
                    "browser": environ.get("SELENIUM_BROWSER"),
                    "url": url,
                    "browser-version": environ.get("SELENIUM_VERSION")
                }
            ]
            if "os" in parsed:
                browsers[0]["os"] = parsed["os"]

        return {
            "server": {
                "host": environ.get("SELENIUM_HOST"),
                "port": environ.get("SELENIUM_PORT"),
                "url": environ.get("SELENIUM_URL"),
                "user": environ.get("SAUCE_USER_NAME"),
                "key": environ.get("SAUCE_API_KEY")
            },
            "browsers": browsers
        }

    @staticmethod
    def config_from_file(absolute_file_path):
        """ Supports reading config a single json file. """
        return json.load(open(absolute_file_path))

    @property
    def xrange(self):
        return xrange(len(self.configs["browsers"]))

    @classmethod
    def from_cli_args(cls):
        import sys
        options = parse_args(sys.argv[1:])
        return cls(options)

    def __init__(self, options):
        """
        Takes the options that are passed into the runner and
        creates a config object that can be referred to throughout
        fattoush.

        :param options: Namespace
        """
        schema_file_name = path.join(path.dirname(__file__),
                                     'file_input_schema.json')

        if options.print_schema:
            with open(schema_file_name) as example:
                print example.read()
            exit(0)
        elif options.print_config:
            file_name = path.join(path.dirname(__file__),
                                  'example_config.json')
            with open(file_name) as example:
                print example.read()
            exit(0)

        self._raw_options = options
        self.parallel = options.parallel

        if options.config is None:
            self.configs = self.config_from_env()
        else:
            self.configs = self.config_from_file(options.config)

        schema = json.load(open(schema_file_name))
        validate(self.configs, schema)

        xunit_filename = ('lettucetests.xml'
                          if options.enable_xunit
                          and options.xunit_file is None
                          else options.xunit_file)

        self.lettuce_options = {
            'base_path': options.base_path,
            'scenarios': options.scenarios,
            'verbosity': options.verbosity,
            'random': options.random,
            'enable_xunit': options.enable_xunit,
            'xunit_filename': xunit_filename,
            'failfast': options.failfast,
            'auto_pdb': options.auto_pdb,
            'tags': ([tag.strip('@') for tag in options.tags]
                     if options.tags else None)
        }

    @property
    def to_dict(self):
        """
        The returned dictionary gives a shallow copy of the data
        required to create a FeatureConfig.
        """
        return {
            'lettuce_options': self.lettuce_options.copy(),
            'config': self.configs.copy()
        }

    def run(self):
        if not self.configs:
            raise IndexError('There are no webdriver configs against '
                             'which to run lettuce.')

        if self.parallel == 'webdriver':
            self._run_parallel()
        else:
            self._run_series()

    def _run_series(self):
        """
        Runs lettuce against each browser configuration one at a time

        :param self: fattoush.config.FattoushConfigGroup
        """
        ex = None
        results = []

        server = self.configs.get("server", {})
        lettuce = self.lettuce_options
        browsers = self.configs["browsers"]

        for (index, browser) in enumerate(browsers):
            try:
                r = run_single(index=index,
                               browser=browser.copy(),
                               server=server.copy(),
                               lettuce=lettuce.copy())
                results.append(r)
            except BaseException as ex:
                print ex
        if ex is not None:
            raise
        return results

    def _run_parallel(self):
        """
        Runs lettuce against all the browser configurations at the same
        time in different processes.

        :param self: fattoush.config.FattoushConfigGroup
        """
        processes = []

        server = self.configs.get("server", {})
        lettuce = self.lettuce_options
        browsers = self.configs["browsers"]

        for (index, browser) in enumerate(browsers):
            process = Process(None, run_single,
                              kwargs=dict(index=index,
                                          browser=browser,
                                          server=server,
                                          lettuce=lettuce))
            processes.append(process)
            process.start()

        for process in processes:
            process.join()


def run_single(index, browser, server, lettuce):
    """
    :param index: int
    :param browser: dict
    :param server: dict
    :param lettuce: dict
    """

    fattoush_config = FattoushConfig(index=index,
                                     browser=browser,
                                     server=server,
                                     lettuce=lettuce)
    world.absorb(fattoush_config, 'fattoush')
    try:
        result = fattoush_config.run()
        return result
    finally:
        world.spew('fattoush')



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

"""
Functions for putting together sauce configuration,
be it from environmental variables or from a JSON string.
"""
import json

from os import path
from lettuce import Runner
from selenium.webdriver import DesiredCapabilities


class FattoushConfig(object):

    def _augment_xunit_filename(self):
        index = self.index
        xunit_file = self.run_args['xunit_filename']

        if xunit_file is None:
            return

        if '.' in path.basename(xunit_file):
            prefix, suffix = path.splitext(xunit_file)
            xunit_file = '{0}_{1}.{2}'.format(prefix, index, suffix)
        else:
            xunit_file = '{0}_{1}'.format(xunit_file, index)
        self.run_args['xunit_filename'] = xunit_file

    def _augment_server(self):
        if self.server is None:
            self.server = {}
        server = self.server
        if "host" not in server:
            server[u"host"] = ("127.0.0.1" if "user" not in server
                               else "ondemand.saucelabs.com")
        if "port" not in server:
            server[u"port"] = ("4444" if "user" not in server
                               else "80")
        if "url" in server and not server["url"]:
            del server["url"]

    def run(self):
        try:
            return Runner(**self.run_args).run()
        except BaseException:
            print "Lettuce raised the following exception:"
            raise

    def _augment_browser(self):
        """
        All parts of the browser config must be a string except for the
        description, which can hold any format.
        """

        def to_string(obj):
            try:
                if isinstance(obj, unicode):
                    return obj
                elif isinstance(obj, list):
                    return ''.join(obj)
            except TypeError:
                return json.dumps(obj)

        if "description" in self.browser:
            self.browser["description"] = to_string(
                self.browser["description"])

    def __init__(self, index, browser, server, lettuce):
        """
        :param index: int
        :param browser: dict
        :param server: dict
        :param lettuce: dict
        """
        self.index = index
        self.server = server
        self.browser = browser.copy()
        self.run_args = lettuce

        self._augment_xunit_filename()
        self._augment_server()

        self._augment_browser()
        self.name = ";".join(('='.join(item) for item in
                              sorted(self.browser.items())
                              if item[0] != "description"))

    @property
    def command_executor(self):
        host = self.server["host"].rstrip('/')

        try:
            protocol, host = host.split("://")
        except ValueError:
            protocol = 'http'

        try:
            host, url_path = host.split(":")
        except ValueError:
            port = self.server["port"]
        else:
            split_path = url_path.split("/")
            port = split_path[0]
            split_path[0] = host
            host = '/'.join(split_path)

        split_path = host.split('/')
        host, split_path = split_path[0], split_path[1:]

        endpoint = "/".join(split_path) if split_path else "wd/hub"

        if "user" in self.server and "key" in self.server:
            return ('{protocol}://{user}:{key}@{host}:{port}/{endpoint}'
                    ).format(protocol=protocol,
                             user=self.server["user"],
                             key=self.server["key"],
                             host=host,
                             port=port,
                             endpoint=endpoint)

        elif "user" in self.server:
            return ('{protocol}://{user}@{host}:{port}/{endpoint}'
                    ).format(protocol=protocol,
                             user=self.server["user"],
                             host=host,
                             port=port,
                             endpoint=endpoint)
        else:
            return ('{protocol}://{host}:{port}/{endpoint}'
                    ).format(protocol=protocol,
                             host=host,
                             port=port,
                             endpoint=endpoint)

    @property
    def browser_lookup(self):
        return {
            "android": DesiredCapabilities.ANDROID,
            "googlechrome": DesiredCapabilities.CHROME,
            "firefox": DesiredCapabilities.FIREFOX,
            "phantom": DesiredCapabilities.PHANTOMJS,
            "htmlunit": DesiredCapabilities.HTMLUNIT,
            "iexploreproxy": DesiredCapabilities.INTERNETEXPLORER,
            "iphone": DesiredCapabilities.IPHONE
        }

    def desired_capabilities(self, name):

        desired = self.browser.copy()
        browser = desired.pop("browser")
        desired["name"] = name

        try:
            capabilities = self.browser_lookup[browser]
        except KeyError:
            capabilities = DesiredCapabilities.CHROME

        capabilities.update(desired)
        return capabilities



# (c) 2014 Mind Candy Ltd. All Rights Reserved.
# Licensed under the MIT License; you may not use this file except in compliance with the License.
# You may obtain a copy of the License at http://opensource.org/licenses/MIT.

from setuptools import setup, find_packages
from os import path
import sys


def read_lines(file_name):
    pth = path.join(path.dirname(__file__), file_name)
    with open(pth) as handle:
        return handle.readlines()

try:
    # If all the prerequisites are installed, we can get the
    # version and description from the package itself.

    # This is mainly intended for package creation and upload,
    # and prevents having to store this information in more than one
    # place.
    sys.path.insert(0, path.join(path.dirname(__file__), 'src'))
    import fattoush
    VERSION = fattoush.VERSION
    LONG_DESCRIPTION = fattoush.__doc__

except:

    # If these pre-requisites are not installed, the version module
    # can still be executed, and the long description is not required
    # for successful installation - mainly I just want to have the
    # correct information in pypi.

    VERSION = "UNKNOWN"
    execfile(path.join(path.dirname(__file__),
                       'src/fattoush/config/version.py'))

    LONG_DESCRIPTION = ""

setup(
    name='Fattoush',
    package_data={'': ['*.txt']},
    author='Alistair Broomhead',
    version=VERSION,
    author_email='alistair.broomhead@mindcandy.com',
    description="A delicious testing framework",
    license='MIT',
    #url='https://github.com/mindcandy/fattoush',
    #download_url='https://github.com/mindcandy/fattoush/zipball/master',
    long_description=LONG_DESCRIPTION,
    packages=find_packages('src'),
    package_dir={'': 'src'},
    install_requires=read_lines("requirements.txt"),
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'fattoush = fattoush.runner.bin:console']
    }
)
